#Stackem tools

import math
import numpy
import pyfits
from astLib.astWCS import WCS
import logging
import scipy.ndimage as ndimage


def logger(level=0):
    logging.basicConfig()

    LOGL = {"0": "INFO",
            "1": "DEBUG",
            "2": "ERROR",
            "3": "CRITICAL"}

    log = logging.getLogger("Stackem")
    log.setLevel(eval("logging."+LOGL[str(level)]))

    return log


def fitsInd(hdr, ctype):
    return filter( lambda ind: hdr["ctype%d"%ind].startswith(ctype), 
            range(1, hdr['naxis']+1) )[0]


def loadFits(imagename):

    with pyfits.open(imagename) as hdu:
        data = hdu[0].data
        hdr = hdu[0].header

    wcs = WCS(hdr, mode="pyfits")

    return data, hdr, wcs


def subregion(data, centre, radius):

    lx, ly = data.shape
    xstart = centre[1] - radius
    xend = centre[1] + radius
    ystart = centre[0] - radius
    yend = centre[0] + radius

    imslice = [slice(None)] * len(data.shape)
    imslice = [slice(xstart, xend),
               slice(ystart, yend)]
                                       
    scube = data[imslice]

    return scube


def sum_region(data, radius):

    w = data.shape[0]
    rad = numpy.linspace(-w/2, w/2, w)
    rad =  numpy.sqrt(rad[numpy.newaxis,:]**2+rad[:,numpy.newaxis]**2)
    mask = rad<=radius

    return (data*mask).sum()


def negnoise(data):
    neg = data[data<0]
    return numpy.concatenate([neg, -neg]).std()


def gauss(x, a0, mu, sigma):
    return  a0*numpy.exp(-(x-mu)**2/(2*sigma**2))


def gauss2d(height, center_x, center_y, width_x, width_y, beta=0):
    """Returns a gaussian function with the given parameters"""

    width_x = float(width_x)
    width_y = float(width_y)
    fct_x = lambda x: (x-center_x)**2/(2*width_x**2)
    fct_y = lambda y: (y-center_y)**2/(2*width_y**2)
    fct_xy = lambda x,y: beta*(x-center_x)*(y-center_x)/(width_x*width_y)

    return lambda x, y: height * numpy.exp(-fct_x(x) - fct_xy(x, y)-fct_y(y))


def elliptical_mask(data, emaj, emin, pa):
        
    px, py = data.shape
    _max = numpy.where(data==data.max())
    ox = _max[0][0] - px/2
    oy = _max[1][0] - py/2

    xx, yy = numpy.ogrid[-px/2:px/2, -py/2:py/2]
    mask0 = ( ((xx-ox)/emaj)**2 + ((yy-oy)/emin)**2 <= 1 )*1
    rmask = ndimage.rotate(mask0, angle=pa, reshape=False)
    
    return rmask!=0 #[(pxn-px)/2:-(pxn-px)/2, (pyn-py)/2:-(pyn-py)/2]


def gauss_weights(data, sigma1, sigma2, mask=None):
    
    px, py = data.shape[-2:]

    x, y = [ numpy.linspace(-a/2, a/2, a) for a in [px, py] ]
    xx, yy = numpy.meshgrid(x, y)

    _max = numpy.where(data==data.max())
    ox = _max[0][0] - px/2
    oy = _max[1][0] - py/2

    mask = 1 if mask is None else mask
    weight = gauss2d(1, ox, oy, sigma1, sigma2)(xx, yy)*mask

    return (data*mask)*weight/weight.sum()



# Image plane stacking tools
# Sphesihle Makhathini <sphemakh@gmail.com>

import matplotlib
matplotlib.use('Agg')

import pylab

import math
import numpy
import pyfits
from Stackem import utils
import sys
from argparse import ArgumentParser
import astLib.astCoords as coords

matplotlib.rcParams.update({'font.size': 18})

def main():

    __version_info__ = (0,0,1)
    __version__ = ".".join( map(str,__version_info__) )

    for i, arg in enumerate(sys.argv):
        if (arg[0] == '-') and arg[1].isdigit(): sys.argv[i] = ' ' + arg

    parser = ArgumentParser(description='Image based stacking tool S Makhathini <sphemakh@gmail.com>')

    add = parser.add_argument
    add("-v","--version", action='version',version='{:s} version {:s}'.format(parser.prog, __version__))

    add("-i", "--image", 
            help="FITS image name")

    add("-c", "--catalog", metavar="CATALOG_NAME:DELIMITER",
            help="Catalog name. Default delimiter is a comma. Format: 'ra,dec,freq' ")

    add("-p", "--prefix", default="gota_stackem_all",
            help="Prefix for output products.")

    add("-w", "--width", type=float, default=0,
            help="For line stacking: Band width [MHz] to sample across frequency (for line stacking). Default is 1."
                 "For continuum stacking: Width (in beams) of subregion to stack. Default is 10 beams")

    add("-vbl", "--vebosity-level", dest="vbl", choices=["0", "1", "2", "3"], default="0",
            help="Verbosity level. 0-> INFO, 1-> DEBUG, 2-> ERROR, 3-> CRITICAL. Default is 0")

    add("-b", "--beam", metavar="BMIN[:BMIN:BPA]",
            help="PSF (a.k.a dirty beam) FWHM in degrees. No default")

    add("-b2p", "--beam2pix", action="store_true",
            help="Do Jy/beam to Jy/pixel conversion")

    add("-L", "--line", action="store_true",
            help="Do line stacking")

    add("-C", "--cont", action="store_true",
            help="Do continuum stacking")

    args = parser.parse_args()


    catalog_string = args.catalog.split(":")
    if len(catalog_string)>1:
        catalgname, delimiter = catalog_string
    else:
        catalogname, delimiter = catalog_string[0], ","

    if args.beam:
        beam = args.beam.split(":")
        if len(beam)==1:
            beam = float(beam[0])
        elif len(beam)==2:
            beam = map(float, beam) + [0]
        else:
            beam = map(float, beam)
    else:
        beam = None

    pylab.clf()

    prefix = args.prefix or "stackem_default"
    pylab.figure(figsize=(15,10))

    if args.line:
        from Stackem import LineStacker

        stack = LineStacker.load(args.image, catalogname, delimiter=delimiter,
                beam=beam, width=args.width, beam2pix=args.beam2pix,
                verbosity=args.vbl)

        stacked_line = stack.stack()*1e6 # convert to uJy
        peak, nu, sigma = gfit_params = stack.fit_gaussian(stacked_line)
        gfit = utils.gauss(range(stack.width), *gfit_params)

        freqs = (numpy.linspace(-stack.width/2, stack.width/2, stack.width)*stack.dfreq - stack.freq0)*1e-9 # convert to GHz

        # plot profile
        pylab.plot(freqs, stacked_line, "r.", label="stacked profile")
        pylab.plot(freqs, gfit, "k-", label="Gaussian fit")
        pylab.grid()
        pylab.xlim(freqs[0], freqs[-1])
        pylab.xlabel("Frequency [GHz]")
        pylab.ylabel("Flux density [mJy/{:s}]".format("beam" if args.beam2pix else "pixel"))
        pylab.legend(loc=1)
        pylab.savefig(prefix+"-line.png")
        pylab.clf()

        # save
        numpy.savetxt(prefix+"-line.txt", stacked_line, delimiter=",")

        with open(prefix+"-line_stats.txt", "w") as info:
            tot = sigma*peak*math.sqrt(2*math.pi)

            stack.log.info("Gaussian Fit parameters.")
            info.write("Gaussian Fit parameters\n")

            info.write("Peak Flux : {:.4g} uJy\n".format(peak))
            stack.log.info("Peak Flux : {:.4g} uJy".format(peak))

            info.write("Integrated Flux : {:.4g} uJy\n".format(tot))
            stack.log.info("Integrated Flux : {:.4g} uJy".format(tot))

            info.write("Profile width : {:.4g} MHz \n".format(sigma*stack.dfreq))
            stack.log.info("Profile width : {:.4g} kHz".format(sigma*stack.dfreq*1e-3))

        stack.log.info("Line stacking ran successfully. Find your outputs at {:s}-line*".format(prefix))

    if args.cont:
        pylab.figure(figsize=(20, 20))
        from Stackem import ContStacker

        stack = ContStacker.load(args.image, catalogname, delimiter=delimiter, 
                beam=beam, beam2pix=args.beam2pix, width=args.width,
                verbosity=args.vbl)

        stacked = stack.stack()

        mask = utils.elliptical_mask(stacked, stack.bmajPix/2., stack.bminPix/2., stack.bpa)

        flux = utils.gauss_weights(stacked, stack.bmajPix/2., stack.bmajPix/2., mask=mask)
        flux = flux.sum()

        rms = utils.negnoise(stacked)

        with open(prefix+"-cont.txt", "w") as cont:
            stack.log.info("Stacked flux: {:.3g} +/- {:.3g} uJy".format(flux*1e6, rms*1e6))
            cont.write("Stacked flux: {:.3g} +/- {:.3g} uJy/beam".format(flux*1e6, rms*1e6))

        # Plot stacked image, and cross-sections

        rotated = numpy.rot90(stacked.T)
        import matplotlib.gridspec as gridspec
        
        gs = gridspec.GridSpec(2, 2,
                               width_ratios = [4,1],
                               height_ratios = [1,4])

        gs.update(wspace=0.05, hspace=0.05)
        ax1 = pylab.subplot(gs[2])
        ax2 = pylab.subplot(gs[0], sharex=ax1)
        ax3 = pylab.subplot(gs[3], sharey=ax1)

        pylab.setp(ax3.get_yticklabels(), visible=False)
        pylab.setp(ax3.get_xticklabels(), rotation=90)
        pylab.setp(ax2.get_xticklabels(), visible=False)

        ax1.imshow(stacked, interpolation="nearest")
        ax1.set_aspect("auto")
        ra0, dec0 = stack.wcs.getCentreWCSCoords()
        ras = numpy.linspace(stack.width/2, -stack.width/2, stack.width)*stack.cell - ra0
        decs = numpy.linspace(-stack.width/2, stack.width/2, stack.width)*stack.cell - dec0

        ax1.set_xticklabels( map(lambda x: coords.decimal2hms(x, ":"), ras) )
        ax1.set_yticklabels( map(lambda x: coords.decimal2dms(x, ":"), decs) )
        ax1.set_xlabel("RA [hms]")
        ax1.set_ylabel("DEC [dms]")
        pylab.setp(ax1.get_xticklabels(), rotation=90)

        ax2.plot(rotated[:,stack.width/2]*1e6)
        ax3.plot(rotated[stack.width/2,:][::-1]*1e6, range(stack.width))
        ax3.set_ylim(0,stack.width-1)
        ax2.set_xlim(0,stack.width-1)
        ax2.set_ylabel("Flux density [uJy/beam]")
        ax3.set_xlabel("Flux density [uJy/beam]")
        pylab.savefig(prefix+"-cont.png")

        # save stacked image as fits
        hdr = stack.hdr
        hdr["crpix1"] = stack.width/2 
        hdr["crpix2"] = stack.width/2 
        pyfits.writeto(prefix+"-cont.fits", stacked, hdr, clobber=True)

        stack.log.info("Continuum stacking ran successfully. Find your outputs at {:s}-line*".format(prefix))



## Image plane continuum stacking script
## Sphesihle Makhathini <sphemakh@gmail.com>

import math
import numpy
import sys
import pylab

from Stackem import utils

from multiprocessing import Process, Manager, Lock
import psutil
manager = Manager()


class load(object):

    def  __init__(self, imagename, catalogname, beam=0, width=None,
                  stokes_ind=0, delimiter=",", beam2pix=False, verbosity=0):
        """ Continuum stacking tool
            imagename: FITS image
            catalogname: List of RA, DEC values. Weights can be added as a 3rd column
            beam: PSF FWHM in degrees
            stokes_id: Index of stokes plane to stack (0,1,2,3) -> (I,Q,U,V)
        """

        self.imagename = imagename
        self.catalogname = catalogname
        self.catalog = numpy.loadtxt(catalogname, delimiter=delimiter)
        self.npos = len(self.catalog)

        if self.catalog.shape[-1]==2:
            self.catalog = numpy.append(self.catalog,
                            numpy.ones([self.npos, 1]), 1)

        self.weights = manager.Value("f",0)

        self.stokes_ind = stokes_ind

        self.log = utils.logger(verbosity)


        self.data, self.hdr, self.wcs = utils.loadFits(imagename)

        # Find restoring beam in FITS header if not specified
        if isinstance(beam, (float, int)):
            if beam==0:
                beam = None
            else:
                self.bmaj = self.bmin = beam
                self.bpa = 0
        elif isinstance(beam, (list, tuple)):
            self.bmaj, self.bmin, self.bpa = beam

        elif beam is None:
            try:
                self.bmaj = self.hdr["bmaj"]
                self.bmin = self.hdr["bmin"]
                self.bpa = self.hdr["bpa"]
            except KeyError: 
                self.log.critical("Beam not specified, and no beam information in FITS header")
        else:
            raise TypeError("Beam must be a list, tuple, int or float")

        self.bmajPix = int(self.bmaj/abs( self.wcs.getXPixelSizeDeg() ) )
        self.bminPix = int(self.bmin/abs( self.wcs.getXPixelSizeDeg() ) )

        self.beamPix = self.bmajPix

        self.ndim = self.hdr["naxis"]

        stokes = self.ndim - utils.fitsInd(self.hdr, "STOKES")
        freq = self.ndim - utils.fitsInd(self.hdr, "FREQ")

        imslice = [slice(None)]*self.ndim
        imslice[stokes] = stokes_ind
        self.data = self.data[imslice].sum(0)
        #raise SystemError((self.data.shape, imslice, freq))

        self.cell = self.wcs.getXPixelSizeDeg()

        self.width = self.beamPix*self.beam if beam else self.beamPix*10
        self.stamps = manager.list([])
        self.beam2pix = beam2pix
        self.track = manager.Value("d",0)
        self.lock = Lock()

    
    def stack(self):

        def worker(ra, dec, weight):

            rapix, decpix = self.wcs.wcs2pix(ra, dec)
            out = utils.subregion(self.data, [rapix, decpix], self.width/2)*weight
            self.lock.acquire()
            if out.shape == (self.width, self.width):
                self.weights.value += weight
                self.stamps.append(out)

            self.track.value += 1
            self.lock.release()

        procs = []
        range_ = range(10, 110, 10)
        print("Progress:"),
        for ra, dec, weight in self.catalog:
            proc = Process(target=worker, args=(ra, dec, weight))
            procs.append(proc)
            proc.start()

            nn = int(self.track.value/float(self.npos)*100)
            if nn in range_:
                print("..{:d}%".format(nn)),
                range_.remove(nn)

        for proc in procs:
            proc.join()

        print("..100%\n")

        stacked = numpy.array(self.stamps).sum(0)/self.weights.value

        return stacked


    def mc_noise_stack(self, imagename, runs=400, stacks=None):
    
        stacks = stacks or xrange(1000, 8000, 1000)

        # get field centre
        cx, cy = self.wcs.wcs2pix(*self.centre)
        self.track.value = 0
    
        def getpos(npos):
            ra = (numpy.random.random(npos) - 0.49)*self.fov/2 + cx
            dec = (numpy.random.random(npos) - 0.49)*self.fov/2 + cy
    
            return numpy.array([ra, dec]).T
    
        noise = manager.list([])
    
        procs = []
        def worker(npos):
            self.track.value += 1
            _noise = numpy.ones(runs)
            for i in range(runs):
                radec = getpos(npos)
                _noise[i] = self.stack()
            
            noise.append([ npos, _noise.std()])
    
        for npos in stacks:
            proc = Process(target=worker, args=(npos,))
            proc.start()
            procs.append(proc)
    
            nn = int(self.track.value/float(len(stacks))*100)
            if nn in range_:
                print("..{:d}%".format(nn)),
                range_.remove(nn)

        for proc in procs:
            proc.join()

        print("..100%\n")
    
        return noise



#!/usr/bin/env python

import os
from setuptools import setup

setup(name="stackem",
    version="0.1.0",
    description="Image plane stacking tools.",
    author="Sphesihle Makhathini",
    author_email="Sphesihle Makhathini <sphemakh@gmail.com>",
    url="https://github.com/sphemakh/Stackem",
    packages=["Stackem"],
    requires=["numpy","matplotlib","scipy", "astLib", "pyfits"],
    scripts=["Stackem/bin/" + i for i in os.listdir("Stackem/bin")], 
    licence="This program should come with the GNU General Public Licence. "\
            "If not, find it at http://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html",
    classifiers=[],
     )



#!/usr/bin/env python

###
# Image based spectral line stacking tool
# S. Makhathini <sphemakh@gmail.com>
###

import math
import numpy 
import pyfits
import math

from scipy.optimize import curve_fit
from multiprocessing import Process, Manager, Lock
from Stackem import utils


class load(object):
    
    def __init__(self, imagename, catalogname, 
                 width=1, beam=None, delimiter=",", 
                 verbosity=0, beam2pix=False):

        self.log = utils.logger(verbosity)

        self.log.info("Laoding Image data and catalog info")

        self.imagename = imagename
        self.catalogname = catalogname
        self.delimiter = delimiter

        # Load data
        self.catalog = numpy.loadtxt(self.catalogname, delimiter=self.delimiter)
        self.nprofs = len(self.catalog)
        
        self.data, self.hdr, self.wcs = utils.loadFits(imagename)

        self.ndim = self.hdr["naxis"]
        self.centre = self.wcs.getCentreWCSCoords()

        self.log.info("Image Centre RA,DEC {:+.3g}, {:+.3g} Deg".format(*self.centre))
        
        cubeslice = [slice(None)]*self.ndim
        if self.ndim >3:
            stokes_ind = self.ndim - utils.fitsInd(self.hdr, "STOKES")
            cubeslice[stokes_ind] = 0

        self.cube = self.data[cubeslice]

        self.profiles = Manager().list([])
        self.weights = Manager().Value("d", 0)

        ind = utils.fitsInd(self.hdr, "FREQ")
        self.crpix = self.hdr["crpix%d"%ind]
        self.crval = self.hdr["crval%d"%ind]
        self.dfreq = self.hdr["cdelt%d"%ind]
        self.freq0 = self.crval + (self.crpix-1)*self.dfreq
        self.nchan = self.hdr["naxis%d"%ind]
        self.width = int(width*1e6/self.dfreq)

        # Find restoring beam in FITS header if not specified
        if isinstance(beam, (float, int)):
            if beam==0:
                beam = None
            else:
                self.bmaj = self.bmin = beam
                self.bpa = 0
        elif isinstance(beam, (list, tuple)):
            self.bmaj, self.bmin, self.bpa = beam

        elif beam is None:
            try:
                self.bmaj = self.hdr["bmaj"]
                self.bmin = self.hdr["bmin"]
                self.bpa = self.hdr["bpa"]
            except KeyError: 
                self.log.critical("Beam not specified, and no beam information in FITS header")
        else:
            raise TypeError("Beam must be a list, tuple, int or float")

        self.bmajPix = int(self.bmaj/abs( self.wcs.getXPixelSizeDeg() ) )
        self.bminPix = int(self.bmin/abs( self.wcs.getXPixelSizeDeg() ) )

        self.beamPix = self.bmajPix
        self.beam2pix = beam2pix
    
        self.excluded = Manager().Value("d",0)
        self.track = Manager().Value("d",0)
        self.lock = Lock()


    def profile(self, radeg, decdeg, cfreq, weight, pid):
        
        rapix, decpix = self.wcs.wcs2pix(radeg, decdeg)

        cfreqPix = int((cfreq - self.freq0)/self.dfreq )

        zstart = cfreqPix - self.width/2
        zend = cfreqPix + self.width/2

        beamPix = self.beamPix

        ystart, yend = (decpix-beamPix/2.), (decpix+beamPix/2.)
        xstart, xend = (rapix-beamPix/2.), (rapix+beamPix/2.)

        self.log.debug("Line profile {:.3f} {:.3f} {:d}-{:d}".format(rapix, decpix, zstart, zend))
        
        pcube = self.cube[zstart:zend, ystart:yend, xstart:xend]

        # Check if this profile is with stacking

        if pcube.shape != (self.width, beamPix, beamPix):
            padz, pady, padx = (0,0), (0,0), (0,0)
            diffx, diffy, diffz = 0, 0, 0

            if pcube.shape[0] != self.width:
                diffz = self.width - pcube.shape[0]
                if cfreqPix < self.cube.shape[0]/2:
                    padz = diffz, 0
                else:
                    padz = 0, diffz

            if pcube.shape[1] != beamPix:
                diffy = beamPix - pcube.shape[1]
                if ystart<0:
                    pady = diffy, 0
                else:
                    pady = 0, diffy
                
            if pcube.shape[2] != beamPix:
                diffx = beamPix - pcube.shape[2]
                if xstart<0:
                    padx = diffx, 0
                else:
                    padx = 0, diffx

            if diffz > self.width/2 or diffx > beamPix/2 or diffy > beamPix/2:
                self.log.debug("Skipping Profile {:d}, its close too an edge (s).".format(pid))
                self.excluded.value += 1
                return
            else:
                npad = padz, pady, padx
                self.log.debug("Profile {:d} is close an edge(s). Padding the exctracted cube by {:s} ".format(pid, repr(npad)))

                pcube = numpy.pad(pcube, pad_width=npad, mode="constant")
        else:
            self.log.debug("Extracting profile {:d}".format(pid))

        self.lock.acquire()
        self.weights.value += weight
        self.track.value += 1
        self.profiles.append(pcube*weight)
        self.lock.release()

        
    def stack(self):

        nprofs = len(self.catalog)
        
        self.log.info("Stacking {:d} line profiles".format(nprofs))

        # Run these jobs in parallel
        procs = []
        range_ = range(10, 110, 10)
        print("Progress:"),
        for i, (ra, dec, cfreq, w, _id) in enumerate(self.catalog):
            proc = Process(target=self.profile, args = (ra, dec, cfreq, w, i) )
            proc.start()
            procs.append(proc)

            nn = int(self.track.value/float(self.nprofs)*100)
            if nn in range_:
                print("..{:d}%".format(nn)),
                range_.remove(nn)
        
        for proc in procs:
            proc.join()

        print("..100%\n")

        self.log.info("Have stackem all")
        self.log.info("{:d} out of {:d} profiles were excluded because they \
were too close to an edge".format(self.excluded.value, nprofs))
        
        stack = numpy.sum(self.profiles, 0)

        mask = utils.elliptical_mask(stack[0], self.bmajPix/2, self.bminPix/2, self.bpa)

        stack = utils.gauss_weights(stack, self.bmajPix/2, self.bminPix/2, mask=mask)
        profile = stack.sum((1,2))/self.weights.value

        return profile


    def fit_gaussian(self, profile):

        nn = len(profile)
        xx = range(nn)
        import scipy.stats as stats
        from scipy.optimize import leastsq

        sigma = 1 #stats.moment(profile, moment=1)
        mu = xx[nn/2]
        peak = profile.max()
        
        def res(p0, x, y):
            peak, mu, sigma = p0
            yf = utils.gauss(x, peak, mu, sigma)
            return y - yf
    
        params = leastsq(res, (peak, mu, sigma), args=(xx, profile))[0]
        
        return params
